import numpy as np
import copy

def loss_regret(g, w, groups):
    ''' Computes the loss regret w.r.t. a grouping of the 
    weight vector using loss gradient.

    Args:
        g (np.array): gradient vector
        w (np.array): weight vector
        groups (numpy.array): mask grouping learners for different 
            delay periods into separate simpilces,
            e.g., np.array([1, 1, 2, 3, 3]) 
            corresponds to models[0:2] playing on one simplex,
            models[2] playing on another, and models[3:] playing 
            on the final simplex. Ususally set to None to perform
            single-simplex hinting.
    '''
    group_keys = list(set(groups))
    regret = np.zeros(g.shape)
    for k in group_keys:
        p_ind = (groups == k)
        regret[p_ind] = np.dot(g[p_ind], w[p_ind]) - g[p_ind] 
    return regret

def normalize_by_groups(w, groups):
    """ Normalize weight vector by groups.

    Args:
        w (np.array): weight vector 
        groups (numpy.array): mask grouping learners for different 
            delay periods into separate simpilces,
            e.g., np.array([1, 1, 2, 3, 3]) 
            corresponds to models[0:2] playing on one simplex,
            models[2] playing on another, and models[3:] playing 
            on the final simplex. Ususally set to None to perform
            single-simplex hinting.
    """
    wout = np.zeros(w.shape)
    group_keys = list(set(groups))
    for k in group_keys:     
        p_ind = (groups == k)                             
        if np.sum(w[p_ind]) > 0.0:
            wout[p_ind] = (w[p_ind]/ np.sum(w[p_ind]))
        else: 
            n_k = sum(p_ind)
            wout[p_ind] = 1./n_k * np.ones(n_k,) # Uniform
    return wout
    


""" Environment class to provide loss objects for online learning.

This abstract base class defines a template for an online learning 
Environment. The Environment object must implement two methods:
    * get_loss: gets a list of (t, loss function) tuples for time t
Users should implement their own derived Environment classes that
overwrite this abstract function and perform whatever 
bookkeeping necessary to provide loss outputs.  

    For example:

    from poold import Environment

    class WeatherEnvironment(Environment):
        def __init__(self, times, delay):
            # Call base class constructor 
            super().__init__(times)

            self.D = delay 

        def get_losses(self, t, os_times=None):
            if os_times is None:
                os_times = range(self.t)
            return [(t_os, self.weather_loss(t_os)) for t_os os_times if t_os < t - self.D]

"""

from abc import ABC, abstractmethod

class Environment(ABC):
    """ Abstract dataset class for online learning """ 
    def __init__(self, times, **kwargs): 
        """ Initialize dataset.

        Args:
            time (list): list of prediction times
        """
        self.times = times 
        self.T = len(times) # algorithm duration 

    @abstractmethod
    def get_losses(self, t, os_times=None, **kwargs):
        """  Get subset of outstanding feedbacks  
        avaliable at time t. If os_times is None, get
        all avaliable feedback.

        Args:
            t: a time representation 
            os_times (list): list of outstanding feedback times
        """
        pass


""" General utility functions. """
import os
import sys
import time
import warnings
import subprocess
import datetime

DATETIME_FORMAT = '%Y%m%d'

def printf(str):
    """Calls print on given argument and then flushes
    stdout buffer to ensure printed message is displayed right away
    """
    print(str, flush=True)

def make_directories(dirname):
    """Creates directory and parent directories with 777 permissions
    if they do not exist
    """
    if dirname != '':
        os.umask(0)
        os.makedirs(dirname, exist_ok=True, mode=0o777)

def make_parent_directories(file_path):
    """Creates parent directories of a given file with 777 permissions
    if they do not exist
    """
    make_directories(os.path.dirname(file_path))


def symlink(src, dest, use_abs_path=False):
    """Symlinks dest to point to src; if dest was previously a symlink,
    unlinks src first

    Args:
      src: source file name
      dest:  target file name
      use_abs_path:  if True, links dest to the absolute path of src
        (useful when src is not expressed relative to dest)
    """
    # n and f flags ensure that prior symlink is overwritten by new one
    if use_abs_path:
        src = os.path.abspath(src)
    cmd = "ln -nsf {} {}".format(src, dest)
    subprocess.call(cmd, shell=True)

def set_file_permissions(file_path, skip_if_exists=False, throw=False,
                         mode=0o777):
    """Set file/folder permissions.

    Parameters
    ----------
    skip_if_exists : boolean
        If True, skips setting permissions if file exists

    throw : boolean
        If True, throws exception if cannot set permissions

    """
    if not skip_if_exists or (skip_if_exists and not os.path.exists(file_path)):
        # Set permissions
        try:
            os.chmod(file_path, mode)
            sh.chown(file_path, group='sched_mit_hill')
        except Exception as err:
            if throw:
                raise err
            else:
                pass

def string_to_dt(string):
    """Transforms string to datetime."""
    return datetime.datetime.strptime(string, DATETIME_FORMAT)

def dt_to_string(dt):
    """Transforms datetime to string."""
    return datetime.datetime.strftime(dt, DATETIME_FORMAT)

def get_dt_range(base_date, days_ahead_start=0, days_ahead_end=0):
    """Lists the dates between (base_date + days_ahead_start), included, and
    (base_date + days_ahead_end), not included.

    Parameters
    ----------
    base_date : datetime
        Reference date for time window.
    days_ahead_start : int
        Time window starts days_ahead_start days from base_date (included).
    days_ahead_end : int
        Time window ends days_ahead_start days from base_date (included).

    Returns
    -------
    List
        Dates in (base_date + days_ahead_start) and (base + days_ahead_end)

    """
    date_start = base_date + datetime.timedelta(days=days_ahead_start)
    days_in_window = days_ahead_end - days_ahead_start

    date_list = [date_start + datetime.timedelta(days=day)
                 for day in range(days_in_window)]
    return(date_list)

def get_current_year():
    """Gets year at the time when the script is run."""
    now = datetime.datetime.now()
    return now.year

class TicToc(object):
    """
    Author: Hector Sanchez
    Date: 2018-07-26
    Description: Class that allows you to do 'tic toc' to your code.

    This class was based the answers that you can find in the next url.
    https://stackoverflow.com/questions/5849800/tic-toc-functions-analog-in-python

    How to use it:

    with TicToc('name'):
      some code....

    or

    t = TicToc('name')
    t.tic()
    some code...
    t.toc()
    print(t.elapsed)

    or

    t = TicToc('name',time.clock) # or any other method.
                                 # time.clock seems to be deprecated
    with t:
      some code....

    or

    t = TicToc()
    t.tic()
    t.tic()
    t.tic()
    t.toc()
    t.toc()
    t.toc()
    print(t.elapsed)

    or

    from src.utils.tictoc import tic,toc

    tic()
    tic()
    toc()
    toc()
    """

    def __init__(self, name='', method='time', nested=False, print_toc=True):
        """
        Args:
        name (str): Just informative, not needed
        method (int|str|ftn|clss): Still trying to understand the default
            options. 'time' uses the 'real wold' clock, while the other
            two use the cpu clock. If you want to use your own method, do it
            through this argument

            Valid int values:
              0: time.time  |  1: time.perf_counter  |  2: time.proces_time

              if python version >= 3.7:
              3: time.time_ns  |  4: time.perf_counter_ns  |  5: time.proces_time_ns

            Valid str values:
              'time': time.time  |  'perf_counter': time.perf_counter
              'process_time': time.proces_time

              if python version >= 3.7:
              'time_ns': time.time_ns  |  'perf_counter_ns': time.perf_counter_ns
              'proces_time_ns': time.proces_time_ns

            Others:
              Whatever you want to use as time.time
        nested (bool): Allows to do tic toc with nested with a single object.
            If True, you can put several tics using the same object, and each toc will
            correspond to the respective tic.
            If False, it will only register one single tic, and return the respective
            elapsed time of the future tocs.
        print_toc (bool): Indicates if the toc method will print the elapsed time or not.
        """
        self.name = name
        self.nested = nested
        self.tstart = None
        if self.nested:
            self.set_nested(True)

        self._print_toc = print_toc

        self._vsys = sys.version_info

        if self._vsys[0] > 2 and self._vsys[1] >= 7:
            # If python version is greater or equal than 3.7
            self._int2strl = ['time', 'perf_counter', 'process_time',
                              'time_ns', 'perf_counter_ns', 'process_time_ns']
            self._str2fn = {'time': [time.time, 's'], 'perf_counter': [time.perf_counter, 's'], 'process_time': [time.process_time, 's'],
                            'time_ns': [time.time_ns, 'ns'], 'perf_counter_ns': [time.perf_counter_ns, 'ns'], 'process_time_ns': [time.process_time_ns, 'ns']}
        elif self._vsys[0] > 2:
            # If python vesion greater than 3
            self._int2strl = ['time', 'perf_counter', 'process_time']
            self._str2fn = {'time': [time.time, 's'], 'perf_counter': [
                time.perf_counter, 's'], 'process_time': [time.process_time, 's']}
        else:
            # If python version is 2.#
            self._int2strl = ['time']
            self._str2fn = {'time': [time.time, 's']}

        if type(method) is not int and type(method) is not str:
            self._get_time = method

        # Parses from integer to string
        if type(method) is int and method < len(self._int2strl):
            method = self._int2strl[method]
        elif type(method) is int and method > len(self._int2strl):
            self._warning_value(method)
            method = 'time'

        # Parses from int to the actual timer
        if type(method) is str and method in self._str2fn:
            self._get_time = self._str2fn[method][0]
            self._measure = self._str2fn[method][1]
        elif type(method) is str and method not in self._str2fn:
            self._warning_value(method)
            self._get_time = self._str2fn['time'][0]
            self._measure = self._str2fn['time'][1]

    def __warning_value(self, item):
        msg = "Value '{0}' is not a valid option. Using 'time' instead.".format(
            item)
        warnings.warn(msg, Warning)

    def __enter__(self):
        if self.nested:
            self.tstart.append(self._get_time())
        else:
            self.tstart = self._get_time()

    def __exit__(self, type, value, traceback):
        self.tend = self._get_time()
        if self.nested:
            self.elapsed = self.tend - self.tstart.pop()
        else:
            self.elapsed = self.tend - self.tstart

        self._print_elapsed()

    def _print_elapsed(self):
        """
        Prints the elapsed time
        """
        if self.name != '':
            name = '[{}] '.format(self.name)
        else:
            name = self.name
        printf('-{0}elapsed time: {1:.3g} ({2})'.format(
            name, self.elapsed, self._measure))

    def tic(self):
        """
        Defines the start of the timing.
        """
        if self.nested:
            self.tstart.append(self._get_time())
        else:
            self.tstart = self._get_time()

    def toc(self, print_elapsed=None):
        """
        Defines the end of the timing.
        """
        self.tend = self._get_time()
        if self.nested:
            if len(self.tstart) > 0:
                self.elapsed = self.tend - self.tstart.pop()
            else:
                self.elapsed = None
        else:
            if self.tstart:
                self.elapsed = self.tend - self.tstart
            else:
                self.elapsed = None

        if print_elapsed is None:
            if self._print_toc:
                self._print_elapsed()
        else:
            if print_elapsed:
                self._print_elapsed()

        # return(self.elapsed)

    def set_print_toc(self, set_print):
        """
        Indicate if you want the timed time printed out or not.
        Args:
          set_print (bool): If True, a message with the elapsed time will be printed.
        """
        if type(set_print) is bool:
            self._print_toc = set_print
        else:
            warnings.warn(
                "Parameter 'set_print' not boolean. Ignoring the command.", Warning)

    def set_nested(self, nested):
        """
        Sets the nested functionality.
        """
        # Assert that the input is a boolean
        if type(nested) is bool:
            # Check if the request is actually changing the
            # behaviour of the nested tictoc
            if nested != self.nested:
                self.nested = nested

                if self.nested:
                    self.tstart = []
                else:
                    self.tstart = None
        else:
            warnings.warn(
                "Parameter 'nested' not boolean. Ignoring the command.", Warning)


class TicToc2(TicToc):
    def tic(self, nested=True):
        """
        Defines the start of the timing.
        """
        if nested:
            self.set_nested(True)

        if self.nested:
            self.tstart.append(self._get_time())
        else:
            self.tstart = self._get_time()


__TICTOC_asdfghh123456789 = TicToc2()
tic = __TICTOC_asdfghh123456789.tic
toc = __TICTOC_asdfghh123456789.toc



""" Visualizations for online learning.

For example:
    import poold
    poold.visualize(learner.history)
    plt.show()

"""
# System imports
import numpy as np
import pandas as pd
import copy
import os

# Plotting imports
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
import matplotlib.gridspec as gridspec
font = {'family' : 'lmodern',
        'weight' : 'normal',
        'size'   : 50}
text = {'usetex' : True}
matplotlib.rc('font',**font)
matplotlib.rc('text',**text)
matplotlib.rcParams['text.latex.preamble']=r"\usepackage{amsmath}"

import seaborn as sns
sns.set(font_scale=1.6)
sns.set_style("white")

def visualize_multiple(experiment_list, style_algs, subset_time=None, filename="temp"):
    num_plots = len(experiment_list)
    fig_r, ax_r = plt.subplots(1, 1, figsize=(10, 4), sharex=False)
    fig_p, ax_p = plt.subplots(1, 1, figsize=(5, 4), sharex=False)
    fig_w, ax_w = plt.subplots(1, num_plots, figsize=(4*num_plots, 4), sharex=False)

    df_losses = None
    for i, (targets, regret_periods, model_alias, history) in enumerate(experiment_list):
        df = visualize(history, regret_periods, targets, model_alias, style_algs, ax=[ax_w[i], ax_r, ax_p], subset_time=subset_time, legend=(i==0))
        if df_losses is None:
            df_losses = copy.copy(df)
        else:
            df_losses = df_losses.merge(df)
    mean_losses = df_losses.mean(axis=0)

    # Save dataframe in latex table format
    fname = f"./eval/losses_{filename}.tex"
    mean_losses.to_latex(fname, float_format="%.3f", longtable=False)

    fig_w.tight_layout()
    fig_w.subplots_adjust(top=0.95, wspace=0, hspace=0)
    fig_r.tight_layout()
    fig_r.subplots_adjust(top=0.95)
    fig_p.tight_layout()
    fig_p.subplots_adjust(top=0.95)

    if not os.path.exists('figs'):
        os.mkdir('figs')
    filename_w = f"./figs/weights_{filename}.pdf"
    filename_r = f"./figs/regret_{filename}.pdf"
    filename_p = f"./figs/params_{filename}.pdf"
    fig_w.savefig(filename_w, bbox_inches='tight')
    fig_r.savefig(filename_r, bbox_inches='tight')
    fig_p.savefig(filename_p, bbox_inches='tight')

    return mean_losses

def visualize(history, regret_periods=None, time_labels=None, model_labels={}, 
    style_algs={}, ax=[None, None, None], params=["lam"], subset_time=None, legend=True):
    """ Visualize online learning losses, weights, and parameters.

    Args:
        history (History): online learning History object
        regret_periods (list[tuple]): list of tuples specifying the start (inclusive) and end
            points (not inclusive) of regret periods
        time_labels (list): list of labels for the time periods
        model_labels (dict): dictionary of model labels
        style_algs (dict): dictionary of model styles
        ax (list[ax]): list of axis objects for plotting the weights, regret, and parameter
            plots respectively.
        params (list[str]): list of parameters to plot
        subset_time (tuple): plot values from times[subset_time[0]:subset_time[1]]
        legend (bool): if True, plot legend.
    """
    times = history.get_times()
    if time_labels is None:
        time_labels = range(len(times))

    if subset_time is not None:
        times = times[subset_time[0]:subset_time[1]]
        time_labels = time_labels[subset_time[0]:subset_time[1]]
    
    if regret_periods is None:
        regret_periods = [(0, len(times))]

    if subset_time is not None:
        subset_regret_periods = []
        for s, e in regret_periods:
            if s in times and e in times:
                subset_regret_periods.append((times.index(s), times.index(e)))
            elif s in times:
                subset_regret_periods.append((times.index(s), len(times)))
            elif e in times:
                subset_regret_periods.append((0, times.index(e)))
        regret_periods = subset_regret_periods

    assert(len(time_labels) == len(times))

    df_losses = pd.DataFrame(columns=history.models+["online_learner"], index=time_labels)
    df_weights = pd.DataFrame(columns=history.models, index=time_labels)
    
    param = history.get_params(0)
    if len(param) > 0:
        param_labels = list(set(params).intersection(set(param.keys())))
        df_params = pd.DataFrame(columns=param_labels, index=time_labels)
    else:
        df_params = None

    for t, time in enumerate(times):
        loss_obj, loss_learner, loss_grad  = history.get_loss(time)
        play_learner = history.get_play(time, return_past=False)
        loss_learner = loss_obj['fun'](w=play_learner)
        params_learner = history.get_params(time)

        loss_all = loss_obj.get('exp', {})
        loss_all['online_learner'] = loss_learner

        # Assign loss and weight dataframe
        df_losses.iloc[t] = loss_all
        df_weights.iloc[t] = dict(zip(history.models, play_learner))
        if df_params is not None:
            df_params.iloc[t] = params_learner

    plot_weights(df_weights, regret_periods, model_labels, style_algs, ax[0], legend, subset_time)
    if not df_losses[history.models].isna().all(axis=None):
        plot_regret(df_losses, regret_periods, model_labels, style_algs, history.models, ax[1], only_learner=True, subset_time=subset_time)
    if df_params is not None:
        plot_params(df_params, regret_periods, model_labels["online_learner"], style_algs, ax[2], subset_time=subset_time)

    return df_losses.rename({"online_learner": model_labels["online_learner"]}, axis=1)

def plot_weights(df, regret_periods, model_labels, style_algs, ax=None, legend=True, subset_time=None):
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(5, 4), sharex=False)

    for m in df.columns:
        alias = model_labels.get(m, m)
        style = style_algs.get(alias, {})
        alias = alias.replace("_", "\_")
        ax.plot(df.index, df[m], label=alias, **style)

    ax.set_title(model_labels.get("online_learner", "Online Learner").replace("_", "\_") + " weights $\mathbf{w}_t$")
    ax.set_ylim([0.0, 1.0])

    if legend:
        handles, labels = ax.get_legend_handles_labels()
        # sort both labels and handles by labels
        labels, handles = zip(*sorted(zip(labels, handles), key=lambda t: t[0]))
        ax.legend(handles, labels,  prop={'size': 15}, ncol=1, loc='best')
    else:
        ax.set_yticks([])
    ticks = ax.get_xticks()
    # ax.set_xticks(ticks[0:-1])

    if subset_time is not None:
        # Date based formatting
        if isinstance(df.index, pd.DatetimeIndex):
            datefmt = mdates.DateFormatter('%b')
            ax.xaxis.set_major_formatter(datefmt)

    plot_time_seperators(regret_periods, df.index, ax)

def plot_regret(df, regret_periods, model_labels, style_algs, input_models, ax=None, only_learner=False, subset_time=None):
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(5, 4), sharex=False)

    # Find best model per regret period
    if only_learner:
        model_list = ["online_learner"]
    else:
        model_list = list(df.columns)

    mean_loss = df.mean(axis=0)
    zeros = pd.Series(0, index=df.index)
    df_regret = pd.DataFrame(columns=model_list, index=df.index)
    for s, e in regret_periods:
        cumsum = df.iloc[s:e].cumsum(axis=0)
        best_model = pd.to_numeric(cumsum[input_models].iloc[-1]).idxmin()
        relative_cumsum = cumsum - cumsum[[best_model]].values
        df_regret.loc[relative_cumsum.index, :] = relative_cumsum[model_list].values

    for m in df_regret.columns:
        alias = model_labels.get(m, m)
        style = style_algs.get(alias, {})
        alias = alias.replace("_", "\_")

        label = f"{alias} \t (RMSE: {mean_loss[m]: .3f})" 
        ax.plot(df_regret.index, df_regret[m], label=label, **style)

    ax.plot(zeros.index, zeros, c='k', linestyle="-")
    ax.set_title("Cumulative regret (RMSE loss)")

    # Sort both labels and handles by labels
    handles, labels = ax.get_legend_handles_labels()
    labels, handles = zip(*sorted(zip(labels, handles), key=lambda t: t[0]))
    ax.legend(handles, labels,  prop={'size': 15}, ncol=2, loc='best')

    plot_time_seperators(regret_periods, df.index, ax)

    if subset_time is not None:
        # Date based formatting
        if isinstance(df.index, pd.DatetimeIndex):
            datefmt = mdates.DateFormatter('%b')
            ax.xaxis.set_major_formatter(datefmt)

def plot_params(df, regret_periods, model_alias, style_algs, ax=None, subset_time=None):
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(5, 4), sharex=False)

    for m in df.columns:
        style = style_algs.get(model_alias, {})
        model_alias = model_alias.replace("_", "\_")
        final_val = df.iloc[-1]["lam"]
        label = f"{model_alias}" + "\t ($\lambda_{T}$ = " + f"{final_val: .3f})"
        ax.plot(df.index, df[m], label=label, **style)

    ticks = ax.get_xticks()
    # ticks[-1] = ticks[-2]
    ax.set_xticks(ticks)

    ax.set_title("Regularization $\lambda_t$")

    handles, labels = ax.get_legend_handles_labels()
    # sort both labels and handles by labels
    labels, handles = zip(*sorted(zip(labels, handles), key=lambda t: t[0]))
    ax.legend(handles, labels,  prop={'size': 15}, ncol=1, loc='best')


    plot_time_seperators(regret_periods, df.index, ax)

    if subset_time is not None:
        # Date based formatting
        if isinstance(df.index, pd.DatetimeIndex):
            datefmt = mdates.DateFormatter('%b')
            ax.xaxis.set_major_formatter(datefmt)

def plot_time_seperators(regret_periods, index, ax):
    ''' Local utiliy function for plotting vertical time seperators '''
    for start, end in regret_periods:
        start_time = index[start]
        if end == len(index):
            end -= 1
        elif end > len(index):
            raise ValueError("Bad time seperator", start, end)
        end_time = index[end]
        ax.axvline(x=start_time, c='k', linestyle='-.', linewidth=1.0)


from setuptools import find_packages, setup

setup(
    name="poold",
    version="0.0.1",
    author="Genevieve Flaspohler and Lester Mackey",
    author_email="geflaspohler@gmail.com",
    description="Python library for Optimistic Online Learning under Delay",
    url="https://github.com/geflaspohler/poold",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
    package_dir={"": "poold"},
    packages=find_packages(where="poold"),
    install_requires=['pandas',
                      'numpy',                     
                      'matplotlib',                     
                      'seaborn',                     
                      'scipy',                     
                      ],
    python_requires=">=3.6",
)



""" Learner class implements online learning algorithms.

This abstract base class defines a template for an online learning 
algorithm. The OnlineLearnere object must implement several methods:
    * update : updates the learner play and parameters 
    * get_params: returns learner parameters
    * reset_params: initializes learner parameters
Several online learning algorithms are implemented as derived
OnlineLearning classes. Learners can be instantiated using
the module's create function.

For example:

    import poold

    models = ["model1", "model2"]
    duration = 20
    learner = poold.create("adahedged", model_list=models, T=duration)

"""
from abc import ABC, abstractmethod
import numpy as np 
import pandas as pd
import copy

from ..utils import loss_regret, normalize_by_groups
from ..utils import History

class OnlineLearner(ABC):
    """ OnlineLearner abstract base class. """    
    def __init__(self, model_list, groups=None, T=None, **kwargs):
        """Initialize online learner. 
        Args:
            model_list (list[str]): list of strings indicating 
                expert model names
                e.g., ["doy", "cfsv2"]
            groups (numpy.array): mask grouping learners for different 
                delay periods into separate simpilces,
                e.g., np.array([1, 1, 2, 3, 3]) 
                corresponds to models[0:2] playing on one simplex,
                models[2] playing on another, and models[3:] playing 
                on the final simplex. Ususally set to None to perform
                single-simplex hinting.
            T (int): > 0, algorithm duration, optional
        """                
        self.t = 0 # current algorithm time
        self.T = T # algorithm horizon
        self.alg = "online_learner"

        # Set up expert models
        self.expert_models = copy.deepcopy(model_list)
        self.d = len(self.expert_models) # number of experts

        # Initialize groups of weight vector into simplices
        if groups is not None:
            self.groups = np.array(groups)
        else:
            self.groups = np.ones(len(self.expert_models),)
        self.group_keys = list(set(self.groups))

        # Create online learning history 
        self.history = History(model_list, default_play=self.init_weights())

        # Oustanding losses 
        self.outstanding = set() 

    @abstractmethod
    def update(self, t_fb, fb, hint, **kwargs):
        """ Algorithm specific parameter updates. If t_fb 
        is None, perform a hint-only parameter update.

        Args:
            t_fb (int): feedback time
            fb (dict): dictionary of play details at 
                feedback time
            hint (np.array): hint vector at time t
        """
        pass

    @abstractmethod
    def get_params(self):
        """ Returns current algorithm hyperparmeters as a dictionary. """
        pass

    def record_losses(self, losses_fb):
        """ Record losses """
        self.history.record_losses(losses_fb)

    def get_play(self, t):
        """ Get play"""
        return self.history.get_play(t)

    def get_grad(self, t):
        """ Get play"""
        return self.history.get_grad(t)

    def update_and_play(self, losses_fb, hint):
        """ Update online learner and generate a new play.
        
        Update weight vector with received feedback
        and any available hints. Update history and return 
        play for time t.
        
        Args:
            losses_fb (list[(int, loss)]): list of 
                (feedback time, loss_feedback) tuples
            hint (dict): hint dictionary of the form:
                {
                    "fun" (callable, optional): function handle for 
                        the hint as a function of play w
                    "grad" (callable): pseudo-gradient vector.
                }
                for the hint pseudoloss at time self.t 
        """
        # Add to set missing feedback
        self.outstanding.add(self.t)

        # Update the history with received losses
        self.history.record_losses(losses_fb)

        # Get hint from input 
        if hint is None:
            # Default of zero optimistic hint
            self.h = np.zeros((self.d,))
        else:
            # Compute loss gradient at current self.w
            self.h = hint['grad'](self.w)  

        # Compute all algorithm updates
        if len(losses_fb) == 0:
            # Hint-only algorithm updates
            self._single_time_update(t_fb=None, hint=self.h)
        else:
            for t_fb, loss_fb in losses_fb:
                self._single_time_update(t_fb=t_fb, hint=self.h)

                # Update history
                self.outstanding.remove(t_fb)

        # Get algorithm parameters
        params = self.get_params()

        # Update play history 
        self.history.record_play(self.t, self.w)
        self.history.record_hint(self.t, self.h)
        self.history.record_params(self.t, params)
        self.history.record_os(self.t, self.outstanding)

        # Update algorithm iteration 
        self.t += 1
        return self.w

    def _single_time_update(self, t_fb, hint): 
        """ Update weight vector with received feedback
        and any available hints.
        
        Args:
            t_fb (int): feedback for round t_fb
            hint (np.array): hint vector
        """
        if t_fb is None:
            self.update(t_fb=None, fb=None, hint=hint)
            return

        # Get play history at time t_fb
        fb = self.history.get(t_fb)

        # Algorithm specific parameter updates
        self.update(t_fb, fb, hint)

    def log_params(self, t):
        """ Return dictionary of algorithm parameters """
        params = self.get_params()
        params['t'] = t
        # Log model weights
        for i in range(self.d):
            params[f"model_{self.expert_models[i]}"] = float(self.w[i])
        return params

    def reset_params(self, T):
        """ Resets algorithm parameters for new duration T. 
        
        Args:
            T (int): > 0, duration
        """
        # Record keeping 
        self.outstanding = set() # currently outstanding feedback

        # Reset algorithm duration
        # self.t = 0 # current algorithm time
        self.T = T # algorithm duration 
        self.h = np.zeros((self.d,)) # last provided hint vector 

    def softmin_by_groups(self, theta, lam):
        """ Return a vector w corresponding to a softmin of
        vector theta with temperature parameter lam

        Args:
            theta (np.array): input vector
            lam (float): temperature parameter 
        """
        # Initialize weight vector
        w = np.zeros((self.d,))

        # Iterate through groupss
        for k in self.group_keys:     
            # Get groups subset
            p_ind = (self.groups == k)                             
            theta_sub = theta[p_ind]
            w_sub = w[p_ind]

            if np.isclose(lam, 0):
                # Return uniform weights over minimizing values
                w_i =  (theta_sub == theta_sub.min()) # get minimum index
                w_sub[w_i] = 1.0  / np.sum(w_i)
            else:
                # Return numerically stable softmin
                minval = np.min(theta_sub)
                w_sub =  np.exp((-theta_sub + minval) / lam) 
                w_sub = w_sub / np.sum(w_sub, axis=None)
            
            w[p_ind] = w_sub
            if not np.isclose(np.sum(w_sub), 1.0):
                raise ValueError(f"Play w does not sum to 1: {w}")

        # Check computation 
        if np.isnan(w).any():
            raise ValueError(f"Update produced NaNs: {w}")
        return w

    def init_weights(self):
        """ Returns uniform initialization weight vector. """          
        w =  np.ones(self.d) / self.d
        w = normalize_by_groups(w, self.groups)
        return w
    
    def get_weights(self):
        ''' Returns dictionary of expert model names and current weights '''
        return dict(zip(self.expert_models, self.w))

    def get_outstanding(self, include=True, **kwargs):
        """ Gets outstanding predictions at time self.t

        Args: 
            include (bool): if True, include current time
                t in oustanding set
        """
        # Add t to oustanding if not already present
        if include: 
            self.outstanding.add(self.t)
        return list(self.outstanding)

    def increment_time(self):
        """ Increment learner time """
        self.t += 1

class AdaHedgeD(OnlineLearner):
    """
    AdaHedgeD module implements delayed AdaHedge and DUB 
    online learning algorithms of Flaspohler et al., "Online 
    Learning with Optimism and Delay" as ODAFTRL on the simplex 
    with different regularization settings.
    """    
    def __init__(self, model_list, groups=None, T=None, reg="adahedged"):
        """ Initializes online_expert 

        Args:
           reg (str): regularization strategy [ "dub | "adahedged" ], 
                for Delayed Upper bound or AdaHedgeD-style 
                adaptive regularization

            Other args defined in OnlineLearner base class.
        """                
        # Base class constructor 
        super().__init__(model_list, groups, T)
        self.alg = "AdaHedgeD"

        # Check and store regulraization 
        supported_reg = ["adahedged", "dub"]
        if reg not in supported_reg:
            raise ValueError(
                f"Unsupported regularizer for AdaHedgeD {reg}.")

        self.reg = reg
        self.reset_params(T)

    def get_params(self):
        """ Returns current algorithm hyperparmeters as a dictionary. """
        return { 
            'lam': self.lam,
            'delta': self.delta,
            'Delta': self.Delta,
        }

    def reset_params(self, T):
        """ Resets algorithm parameters for new duration T. 
        
        Args:
            T (int): > 0, duration
        """
        # Base class reset 
        super().reset_params(T)
        
        # Initialize play
        self.w = self.init_weights() # uniform weights

        #  Algorithm parameters 
        self.theta = np.zeros((self.d, )) # dual-space parameter 
        self.lam = 0.0 # time varying regularization

        # Regularization parameters
        self.alpha = np.log(self.d) # alpha parameter
        self.at_max = 0.0 # running max of a_t terms for DUB
        self.at_prev = [] # history of a_t terms
        self.delta = 0.0 # per-iteration increase in step size
        self.Delta = 0.0 # cummulative sum of a_t^2 + 2b_t terms for DUB

    def update(self, t_fb, fb, hint):
        """ Algorithm specific parameter updates. If t_fb 
        is None, perform a hint-only parameter update 

        Args:
            t_fb (int): feedback time
            fb (dict): dictionary of play details at 
                feedback time
            hint (np.array): hint vector at time t
        """
        # Hint only update
        if t_fb is None:
            self.w = self.softmin_by_groups(self.theta + hint, self.lam)
            return

        # Get feedback gradient
        g_fb = fb['g']

        # Update dual-space parameter value with standard 
        # gradient update, sum of gradients
        self.theta = self.theta + g_fb 

        # Update regularization
        assert("lam" in fb["params"])
        if self.reg == "adahedged":
            self.lam, self.delta  = self.get_reg(
                g_fb, fb["w"], fb["h"],
                fb["g_os"], fb["params"]["lam"])
        elif self.reg == "dub":
            self.lam, self.delta  = self.get_reg_dub(
                g_fb, fb["w"], fb["h"],
                fb["g_os"], fb["D"])
        else:
            raise ValueError(f"Unrecognized regularizer {self.reg}")

        # Update expert weights 
        self.w = self.softmin_by_groups(self.theta + hint, self.lam)

    def get_reg(self, g_fb, w_fb, hint_fb, g_os, lam_fb):
        """ Returns an updated AdaHedgeD-style regularizer
            g_fb (numpy.array): most recent feedback gradient t-D
            w_fb (numpy.array): play at time t-D
            hint_fb (numpy.array): hint at t-D
            g_os (numpy.array): sum of gradients outstanding at time t-D
            lam_fb (float): value of regularizer at t-D
        """
        # Get delta value
        delta = self.get_delta(g_fb, w_fb, hint_fb, g_os, lam_fb)

        # Update regularization
        eta = self.lam + delta / self.alpha 
        return eta, delta

    def get_delta(self, g_fb, w_fb, hint_fb, g_os, lam_fb):
        """ Computes change to AdaHedgeD-style regularizer.
            g_fb (numpy.array): most recent feedback gradient t-D
            w_fb (numpy.array): play at time t-D
            hint_fb (numpy.array): hint at t-D
            g_os (numpy.array): sum of gradients outstanding at time t-D
            lam_fb (float): value of regularizer at t-D
        """
        # Compute the Be-The-Regularized-Leader Solution using
        # \lam_{t-D} and g_{1:t-D}
        w_btrl = self.softmin_by_groups(self.theta, lam_fb)

        # Compute drift delta
        delta_drift = np.dot(g_fb, w_fb - w_btrl)

        # Compute auxiliary regret delta for the hint term
        if np.isclose(lam_fb, 0):
            delta_aux = np.dot(self.theta, w_fb) - np.min(self.theta)
        else:
            delta_aux = self.get_aux_delta(lam_fb, hint_fb, g_os, w_fb)

        # Compute the auxiliary regret delta for the mixed term
        g_norm = np.linalg.norm(g_fb, ord=np.inf) # gradient norm g_t-D
        err_norm = np.linalg.norm(hint_fb - g_os, ord=np.inf) # hint error norm h_t-D - g_{t-2D:t-D}
        sigma = min(1.0, g_norm / err_norm) # sigma weighing between hint and ground truth
        h_hat = g_os + sigma * (hint_fb - g_os) # interpolated hint
        w_hat = self.softmin_by_groups(self.theta-g_os+h_hat, lam_fb) # play using interpolated hint

        if np.isclose(lam_fb, 0):
            delta_hat = np.dot(self.theta, w_hat) - np.min(self.theta) + np.dot(g_fb, w_fb - w_hat)
        else:
            delta_hat = self.get_aux_delta(lam_fb, h_hat, g_os, w_hat)
            delta_hat += np.dot(g_fb, w_fb - w_hat)

        # Compute final delta term
        delta = max(min([delta_drift, delta_aux, delta_hat]), 0.0)
        return delta

    def get_aux_delta(self, lam_fb, hint, g_os, w):
        """ The delta value for lam != 0 """
        g_diff = hint - g_os
        maxval = np.max(g_diff[w != 0.0])

        delta = lam_fb * \
            np.log(np.sum(w * np.exp((g_diff - maxval) / lam_fb))) + \
            np.dot(-g_diff, w) + maxval

        return delta

    def get_reg_dub(self, g_fb, w_fb, hint_fb, g_os, D):
        """ Delayed upper bound regularization value
            g_fb (numpy.array): most recent feedback gradient t-D
            hint_fb (numpy.array): hint at t-D
            g_os (numpy.array): sum of gradients outstanding at time t-D
            D (int): length of delay
        """
        # Get a_t and b_t upper bounds
        a_t = self.get_at_bound(g_fb, hint_fb, g_os)
        b_t = self.get_bt_bound(g_fb, hint_fb, g_os)

        # TODO: can modify this to a max length queue to avoid
        # saving the entire history of a_ts
        self.at_prev.append(copy.copy(a_t))
        at_delay_sum = sum(self.at_prev[-D:-1])

        # Update max a_t term
        self.at_max = np.max([self.at_max, at_delay_sum])

        # Update delta term
        delta = a_t**2 + 2*self.alpha*b_t
        self.Delta += delta

        eta = (2*self.at_max + np.sqrt(self.Delta)) / self.alpha

        # Ensure monotonic increases
        return eta, delta 

    def get_at_bound(self, g_fb, hint_fb, g_os):
        """ Get bound on the value of a_t terms, assume diam(W) = 2

        Args:
            g_fb (numpy.array): most recent feedback gradient t-D
            hint_fb (numpy.array): hint at t-D
            g_os (numpy.array): sum of gradients outstanding at time t-D
        """
        g_norm = np.linalg.norm(g_fb, ord=np.inf)
        err_norm = np.linalg.norm(hint_fb - g_os, ord=np.inf)
        return 2*np.min([g_norm, err_norm])

    def get_bt_bound(self, g_fb, hint_fb, g_os):
        """ Get bound on the value of b_t terms, assume diam(W) = 2

        Args:
            g_fb (numpy.array): most recent feedback gradient t-D
            hint_fb (numpy.array): hint at t-D
            g_os (numpy.array): sum of gradients outstanding at time t-D
        """
        g_norm = np.linalg.norm(g_fb, ord=np.inf)
        err_norm = np.linalg.norm(hint_fb - g_os, ord=np.inf)
        return 0.5*(err_norm**2) - 0.5*np.max([0, err_norm - g_norm])**2

class DORM(OnlineLearner):
    """
    DORM module implements delayed optimistc regret matching 
    online learning algorithm of Flaspohler et al., "Online 
    Learning with Optimism and Delay" with q=2.
    """    
    def __init__(self, model_list, groups=None, T=None):
        """ Initializes online_expert 

        Args:
            Args defined in OnlineLearner base class.
        """                
        # Base class constructor 
        super().__init__(model_list, groups, T)
        self.alg = "DORM"

        #  Initialize learner
        self.reset_params(T)

    def get_params(self):
        """ Returns current algorithm hyperparmeters as a dictionary. """
        return {}

    def reset_params(self, T):
        """ Resets algorithm parameters for new duration T. 
        
        Args:
            T (int): > 0, duration
        """
        # Base class reset 
        super().reset_params(T)

        # Initialize play
        self.w = self.init_weights() # uniform weights

        #  Algorithm parameters 
        self.regret = np.zeros((self.d,)) # cumulative regret vector 

    def update(self, t_fb, fb, hint):
        """ Algorithm specific parameter updates. If t_fb 
        is None, perform a hint-only parameter update 

        Args:
            t_fb (int): feedback time
            fb (dict): dictionary of play details at 
                feedback time
            hint (np.array): hint vector at time t
        """
        # Hint only update
        if t_fb is None:
            regret_pos = np.maximum(0, hint)
            self.w = normalize_by_groups(regret_pos, self.groups)
            return 

        # Update dual-space parameter value with standard 
        # regret gradient update, sum of gradients
        assert("w" in fb)
        assert("g" in fb)
        g_fb = fb['g'] # get feedback gradient
        w_fb = fb["w"] # get feedback play 
        regret_fb = loss_regret(g_fb, w_fb, self.groups) # compute regret w.r.t. groups 
        self.regret = self.regret + regret_fb 

        # Update regret
        regret_pos = np.maximum(0, self.regret + hint)

        # Update expert weights 
        self.w = normalize_by_groups(regret_pos, self.groups)

class DORMPlus(OnlineLearner):
    """
    DORMPlus module implements delayed optimistc regret matching+
    online learning algorithm of Flaspohler et al., "Online 
    Learning with Optimism and Delay" with q=2.
    """    
    def __init__(self, model_list, groups=None, T=None):
        """ Initializes online_expert 

        Args:
            Args defined in OnlineLearner base class.
        """                
        # Base class constructor 
        super().__init__(model_list, groups, T)
        self.alg = "DORMPlus"

        #  Initialize learner
        self.reset_params(T)

    def get_params(self):
        """ Returns current algorithm hyperparmeters as a dictionary. """
        return {}

    def reset_params(self, T):
        """ Resets algorithm parameters for new duration T. 
        
        Args:
            T (int): > 0, duration
        """
        # Base class reset 
        super().reset_params(T)

        #  Algorithm parameters 
        self.w = self.init_weights() # uniform weights, initial play
        self.p = np.zeros((self.d,)) # must initialize initial pseudo-play to zero
        self.hint_prev = np.zeros((self.d,)) # past hint

    def update(self, t_fb, fb, hint):
        """ Algorithm specific parameter updates. If t_fb 
        is None, perform a hint-only parameter update 

        Args:
            t_fb (int): feedback time
            fb (dict): dictionary of play details at 
                feedback time
            hint (np.array): hint vector at time t
        """
        # Hint only update
        if t_fb is None:
            self.p = np.maximum(0, self.p + hint - self.hint_prev)
            self.w = normalize_by_groups(self.p, self.groups)
            self.hint_prev = copy.deepcopy(hint)
            return 

        # Update dual-space parameter value with standard 
        # regret gradient update, sum of gradients
        assert("w" in fb)
        assert("g" in fb)
        w_fb = fb["w"]
        g_fb = fb["g"]
        regret_fb = loss_regret(g_fb, w_fb, self.groups) # compute regret w.r.t. groups 

        # Update psuedo-play 
        self.p = np.maximum(0, self.p + regret_fb + hint - self.hint_prev)

        # Update expert weights 
        self.w = normalize_by_groups(self.p, self.groups) 

        # Update previous hint
        self.hint_prev = copy.deepcopy(hint)

class ReplicatedOnlineLearner(object):
    """ ReplicatedOnlineLearner class. """    
    def __init__(self, learner, replicates=1):
        """Initialize online learner. 
        Args:
            learner (OnlineLearner): an online learning object
            replicates (int): number of replicates
        """              
        self.base_learner = copy.deepcopy(learner) 
        self.learner_queue = [copy.deepcopy(learner) for i in range(replicates)]
        self.t = 0 # current algorithm time
        self.c = 0 # current learner count
        self.t_to_c = {0: 0} # time to learner count mapping
        self.d = self.base_learner.d
        self.groups = self.base_learner.groups
        self.reps = replicates
        self.alg = self.base_learner.alg

        # Create online learning history 
        self.history = History(learner.expert_models, default_play=learner.init_weights())

    def get_params(self, t):
        """ Returns current algorithm hyperparmeters as a dictionary. """
        return self.learner_queue[self.t_to_c[t] % self.reps].get_params()

    def update_and_play(self, losses_fb, hint):
        """ Update online learner and generate a new play.
        
        Update weight vector with received feedback
        and any available hints. Update history and return 
        play for time t.
        
        Args:
            losses_fb (list[(int, loss)]): list of 
                (feedback time, loss_feedback) tuples
            hint (dict): hint dictionary of the form:
                {
                    "fun" (callable, optional): function handle for 
                        the hint as a function of play w
                    "grad" (callable): pseudo-gradient vector.
                }
                for the hint pseudoloss at time self.t 
        """
        # Update the history with received losses
        self.history.record_losses(losses_fb)

        w = self.learner_queue[self.c % self.reps].update_and_play(losses_fb, hint)

        # Get algorithm parameters
        params = self.learner_queue[self.c % self.reps].get_params()

        # Update play history 
        self.history.record_play(self.t, w)
        self.history.record_params(self.t, params)

        # Increment time for the rest of the learners 
        for i, learner in enumerate(self.learner_queue):
            if i != self.c % self.reps:
                learner.t += 1

        # Log and update counters
        self.t += 1
        self.c += 1
        self.t_to_c[self.t] = self.c
        # Update algorithm iteration 
        return w

    def log_params(self, t):
        """ Return dictionary of algorithm parameters """
        learner = self.learner_queue[self.t_to_c[t] % self.reps]
        params = learner.get_params()
        params['t'] = t
        # Log model weights
        for i in range(self.d):
            params[f"model_{learner.expert_models[i]}"] = float(learner.w[i])
        return params

    def reset_params(self, T):
        """ Resets algorithm parameters for new duration T. 
        
        Args:
            T (int): > 0, duration
        """
        for i in range(self.reps):
            self.learner_queue[i].reset_params(T)

    def increment_time(self):
        """ Increment learner time """
        self.t += 1
        self.t_to_c[self.t] = self.c
        for i in range(self.reps):
            self.learner_queue[i].increment_time()

    def record_losses(self, losses_fb):
        """ Record losses """
        self.history.record_losses(losses_fb)
        for t_fb, loss_fb in losses_fb:
            self.learner_queue[self.t_to_c[t_fb] % self.reps].record_losses([(t_fb, loss_fb)])

    def get_play(self, t):
        """ Get play"""
        return self.learner_queue[self.t_to_c[t] % self.reps].get_play(t)

    def get_grad(self, t):
        """ Get grad """
        return self.learner_queue[self.t_to_c[t] % self.reps].get_grad(t)

    def get_outstanding(self, include=True, all_learners=False):
        """ Gets outstanding predictions at time self.t

        Args: 
            include (bool): if True, include current time
                t in oustanding set
        """
        # Add t to oustanding if not already present
        learner = self.learner_queue[self.c % self.reps]
        if include: 
            learner.outstanding.add(self.t)

        # Get either current learner outstanding set or
        # oustanding set from all learners
        outstanding = set()
        if all_learners:
            for i in range(self.reps):
                outstanding = outstanding.union(self.learner_queue[i].outstanding)
        else:
            outstanding = learner.outstanding

        return list(outstanding)
